# Copyright 2015, 2016 Reahl Software Services (Pty) Ltd. All rights reserved.
# -*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
""".. versionadded:: 3.2

Reahl uses CSS and JavaScript from other projects. The CSS and JavaScript world
has its own ecosystem of tools that allow you to do various things, for example:

 - Compile your own customised CSS using a CSS preprocessor like `Sass <http://sass-lang.com/>`_\.
 - Manage the versions and interdependencies of such projects.
 - Some code is hosted on Content Delivery Networks (CDNs).
 - Some code is minified.

This module is a start towards supporting such "front-end frameworks"
which Reahl depends upon. Currently, Reahl merely includes the correct
versions of the front-end libraries it needs and a user is not able to
customise these in any way. In the future we hope to extend this
infrastructure to be able to make use of everything the JavaScript /
CSS tool ecosystem offers.

Reahl also contains its own JavaScript and CSS code. Where we do this,
our own code is also managed using this same infrastructure.

If you write your own Widgets that make use of CSS or JavaScript,
creating a :class:`Library` is the way to do it going forward.

"""

import itertools
from collections import OrderedDict

from reahl.component.exceptions import ProgrammerError
from reahl.component.context import ExecutionContext
from reahl.web.fw import PackagedFile, ConcatenatedFile

class LibraryIndex(object):
    """An ordered collection of :class:`Library` instances.

    A LibraryIndex instance is available in the configuration as
    `web.config.frontend_libraries`. All the :class:`Library`
    instances contained in `web.config.frontend_libraries` are
    included on any :class:`~reahl.web.ui.HTML5Page`.

    Libraries in a LibraryIndex are included on a page in the order
    that they appear in the LibraryIndex.

    :param libraries: :class:`Library` instances (in order) to initialise this LibraryIndex with.

    """
    def __init__(self, *libraries):
        self.libraries_by_name = OrderedDict()
        for library in libraries:
            self.add(library)

    def clear(self):
        """Removes all Libraries from the list."""
        self.libraries_by_name.clear()

    def add(self, library):
        """Adds a Library to the end of the list."""
        self.libraries_by_name[library.name] = library
        return library

    def __contains__(self, name):
        """An implementation of the `in` operator, so that one can ask whether a library with given name is in this index.

        :param name: The unique name of the Library to check for.
        """
        return name in self.libraries_by_name

    def packaged_files(self):
        return [i for i in itertools.chain(*[library.packaged_files() for library in self])]

    def __iter__(self):
        return iter(self.libraries_by_name.values())
        
    def use_deprecated_yui(self):
        """A convenience method for modifying the standard LibraryIndex to
        switch on the deprecated yuigridscss Library.
        """
        if 'pure' in self:
            del self.libraries_by_name['pure']
        if 'yuigridscss' not in self:
            self.add(YuiGridsCss())

    def enable_experimental_bootstrap(self):
        """A convenience method for modifying the standard LibraryIndex to
        switch on experimental Bootstrap Library and its
        dependencies.
        """
        if 'pure' in self:
            del self.libraries_by_name['pure']
        self.add(Tether())
        self.add(Bootstrap4())
        self.add(ReahlBootstrap4Additions())
        

class Library(object):
    """A frontend-library: a collection of CSS and JavaScript code that can be used with Reahl.

    To create your own Library, subclass from this class and set its
    attributes to indicate which files to include as part of it and
    where to find them.

    To use a library, add it to the `web.config.frontend_libraries`
    config setting. The CSS and JavaScript files of all such
    configured libraries are automatically included in any
    :class:`~reahl.web.ui.HTML5Page`.

    :param name: A unique name for this Library.
    """
    def __init__(self, name):
        self.name = name  #: The unique name of this Library
        self.egg_name = 'reahl-web'  #: The component (egg) that contains the files of this library
        self.shipped_in_directory = ''  #: The directory that contains the files of this library (relative to the egg root)
        self.files = []   #: The JavaScript and CSS files that form part of this library (relative to the `shipped_in_directory`)

    def packaged_files(self):
        return [PackagedFile(self.egg_name, self.shipped_in_directory, file_name) 
                for file_name in self.files]

    def packaged_files222xxxxxFIXMEIAMDEADButPossibleAlternative(self):
        exposed_files = []
        js_files_to_include = [PackagedFile(self.egg_name, self.shipped_in_directory, file_name) 
                               for file_name in self.files
                               if file_name.endswith('.js')]
        if js_files_to_include:
            exposed_files.append(ConcatenatedFile('%s.js' % self.name, js_files_to_include))

        css_files_to_include = [PackagedFile(self.egg_name, self.shipped_in_directory, file_name) 
                               for file_name in self.files
                               if file_name.endswith('.css')]
        if css_files_to_include:
            exposed_files.append(ConcatenatedFile('%s.css' % self.name, css_files_to_include))
        return exposed_files


    def files_of_type(self, extension):
        return [f for f in self.files if f.endswith(extension)]

    def header_only_material(self, rendered_page):
        result = ''
        for file_name in self.files_of_type('.css'):
            result += '\n<link rel="stylesheet" href="/static/%s" type="text/css">' % file_name
        return result

    def footer_only_material(self, rendered_page):
        result = ''
        for file_name in self.files_of_type('.js'):
            result += '\n<script type="text/javascript" src="/static/%s"></script>' % file_name
        return result


    
class JQuery(Library):
    """Version 1.11.2 of `JQuery <https://jquery.com>`_.

    This Library also includes a number of plugins we use internally:

    =================== ==================================
     Plugin              Version
    =================== ==================================
     jquery-migrate      1.2.1
     jquery.cookie       1.0
     jquery.metadata     2.1
     jquery.validate     1.10.0 (a heavily modified version)
     jquery.ba-bbq       1.2.1
     jquery.blockUI      2.70.0
     jquery.form         3.14
    =================== ==================================
    """
    def __init__(self):
        super(JQuery, self).__init__('jquery')
        self.files = ['jquery-1.11.2/jquery-1.11.2.js',
                      'jquery-1.11.2/jquery-1.11.2.min.map']
        self.shipped_in_directory = '/reahl/web/static'
        for i in ['jquery-migrate-1.2.1.js',
                  'jquery.cookie-1.0.js', 
                  'jquery.metadata-2.1.js',
                  'jquery.validate-1.10.0.modified.js',
                  'jquery.ba-bbq-1.2.1.js',
                  'jquery.blockUI-2.70.0.js',
                  'jquery.form-3.14.js']:
            self.add_shipped_plugin('jquery/%s' % i)

    def add_shipped_plugin(self, file_name):
        self.files.append(file_name)

    def document_ready_material(self, rendered_page):
        result = '\n<script type="text/javascript">\n'
        result += 'jQuery(document).ready(function($){\n'
        result += '$(\'body\').addClass(\'enhanced\');\n'
        js = set()
        js.update(rendered_page.get_js())
        for item in sorted(js):
            result += item+'\n'
        result += '\n});'
        result += '\n</script>\n'
        return result

    def footer_only_material(self, rendered_page):
        result = super(JQuery, self).footer_only_material(rendered_page)
        #from http://ryanpricemedia.com/2008/03/19/jquery-broken-in-internet-explorer-put-your-documentready-at-the-bottom/
        result += self.document_ready_material(rendered_page)
        return result
        
        

        
class JQueryUI(Library):
    """A heavily customised subset of version 1.10.3 of `JQuery UI <https://jqueryui.com>`_.
    
    .. warning:: 
    
       This Library will be trimmed in future to only contain
       the `Widget Factory <http://api.jqueryui.com/jQuery.widget/>`_.
    """
    def __init__(self):
        super(JQueryUI, self).__init__('jqueryui')
        self.shipped_in_directory = '/reahl/web/static'
        self.files = ['jquery-ui-1.10.3.custom.js']



class Pure(Library):
    """Version 0.5.0 of `Pure UI <http://purecss.io>`_.
    
    .. warning:: This Library will be removed in future versions as part of our intended move to Bootstrap.
    """
    def __init__(self):
        super(Pure, self).__init__('pure')
        self.shipped_in_directory = '/reahl/web/static/pure-release-0.5.0'
        self.files = ['base.css', 'grids.css', 'grids-responsive.css']


class YuiGridsCss(Library):
    """A customised version of `Yui Grids CSS <http://yuilibrary.com/yui/docs/cssgrids/>`_.

    .. deprecated:: 3.1

    .. warning:: 

       This Library is not enabled by default. It is currently still
       here to provide backwards compatibility. Note that it will be
       removed in future versions as part of our intended move to
       Bootstrap.

    """
    @classmethod
    def is_enabled(cls):
        frontend_libraries = ExecutionContext.get_context().config.web.frontend_libraries
        return 'yuigridscss' in frontend_libraries

    @classmethod
    def check_enabled(cls, calling_object):
        if not cls.is_enabled():
            raise ProgrammerError('YuiGridsCss not enabled in current configuration. For %s to work, add the line: \n'\
                                  '"web.frontend_libraries.use_deprecated_yui()"\n in your web.config.py'\
                                   % calling_object.__class__.__name__)

    def __init__(self):
        super(YuiGridsCss, self).__init__('yuigridscss')
        self.shipped_in_directory = '/reahl/web/static'
        self.files = ['reset-fonts-grids.css']


class HTML5Shiv(Library):
    """Version 3.6.3 of `html5shiv <https://github.com/aFarkas/html5shiv>`_.
    
    """
    def __init__(self):
        super(HTML5Shiv, self).__init__('html5shiv')
        self.shipped_in_directory = '/reahl/web/static'
        self.files = ['html5shiv-printshiv-3.6.3.js']

    def footer_only_material(self, rendered_page):
        # From: http://remysharp.com/2009/01/07/html5-enabling-script/ 
        result  = '\n<!--[if lt IE 9]>'
        result  += super(HTML5Shiv, self).footer_only_material(rendered_page)
        result  += '<![endif]-->'
        return result


class IE9(Library):
    """Version 2.1(beta4) of `IE9.js <https://code.google.com/archive/p/ie7-js/>`_.
    """
    def __init__(self):
        super(IE9, self).__init__('IE9')
        self.shipped_in_directory = '/reahl/web/static'
        self.files = ['IE9.js']

    def footer_only_material(self, rendered_page):
        # From: http://code.google.com/p/ie7-js/ 
        # Not sure if this does not perhaps interfere with Normalize reset stuff? 
        result  = '\n<!--[if lte IE 9]>'
        result  += super(IE9, self).footer_only_material(rendered_page)
        result  += '<![endif]-->'
        return result


class Reahl(Library):
    """JavaScript and CSS that is part of Reahl itself.
    """
    def __init__(self):
        super(Reahl, self).__init__('reahl')
        self.shipped_in_directory = '/reahl/web'
        self.files = ['reahl.hashchange.js',
                      'reahl.ajaxlink.js',
                      'reahl.textinput.js',
                      'reahl.runningonbadge.css']

    def header_only_material(self, rendered_page):
        return super(Reahl, self).header_only_material(rendered_page) + \
               '\n<link rel="stylesheet" href="/static/reahl.css" type="text/css">' 

    def footer_only_material(self, rendered_page):
        return super(Reahl, self).footer_only_material(rendered_page) + \
               '\n<script type="text/javascript" src="/static/reahl.js"></script>'


class Holder(Library):
    """Version 2.9.0 of `Holder <http://imsky.github.io/holder/>`_.
    """
    def __init__(self):
        super(Holder, self).__init__('holder')
        self.shipped_in_directory = '/reahl/web/holder'
        self.files = ['holder-2.9.0.js']


class Bootstrap4(Library):
    """Version 4.0.0 alpha 2 of `Bootstrap <http://getbootstrap.com/>`_.
    """
    def __init__(self):
        super(Bootstrap4, self).__init__('bootstrap4')
        self.shipped_in_directory = '/reahl/web/static'
        self.files = [
                      'bootstrap-4.0.0-alpha.2/css/bootstrap.css',
                      'bootstrap-4.0.0-alpha.2/css/bootstrap.css.map',
                      'bootstrap-4.0.0-alpha.2/js/bootstrap.js'
                      ]


    def header_only_material(self, rendered_page):
        return '<meta http-equiv="x-ua-compatible" content="ie=edge">'\
               '<meta name="viewport" content="width=device-width, initial-scale=1">' +\
               super(Bootstrap4, self).header_only_material(rendered_page) 



class ReahlBootstrap4Additions(Library):
    """Reahl specific JavaScript and CSS for use with :class:`Bootstrap4`.
    """
    def __init__(self):
        super(ReahlBootstrap4Additions, self).__init__('bootstrap4.reahladditions')
        self.shipped_in_directory = '/reahl/web/bootstrap'
        self.files = [
                      'reahl.bootstrapform.js',
                      'reahl.pagination.js',
                      'reahl.files.js',
                      'reahl.files.css',
                      'reahl.bootstrappopupa.js',
                      'reahl.carousel.css',
                      'reahl.bootstrapcueinput.js',
                      'reahl.bootstrapfileuploadli.js',
                      'reahl.bootstrapfileuploadpanel.js',
                      'reahl.datatable.css'                      
                      ]

class Tether(Library):
    """Version 1.1.1 of `Tether <http://github.hubspot.com/tether/>`_.
    """
    def __init__(self):
        super(Tether, self).__init__('tether')
        self.shipped_in_directory = '/reahl/web/static'
        self.files = [
                      'tether.1.1.1/css/tether.css',
                      'tether.1.1.1/css/tether-theme-arrows.css',
                      'tether.1.1.1/css/tether-theme-arrows-dark.css',
                      'tether.1.1.1/css/tether-theme-basic.css',
                      'tether.1.1.1/js/tether.js'
                      ]









# Copyright 2016 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

""".. versionadded:: 3.2

A Bootstrap Navbar is a header for a web application with all manner
of useful content, including a navigation menu.

"""
from __future__ import print_function, unicode_literals, absolute_import, division

import six

from reahl.component.exceptions import arg_checks, IsInstance, ProgrammerError

from reahl.web.fw import Layout
from reahl.web.ui import Url, HTMLElement, HTMLWidget, HTMLAttributeValueOption
from reahl.web.bootstrap.ui import Div, Nav, A, TextNode
from reahl.web.bootstrap.forms import Form
import reahl.web.bootstrap.navs
from reahl.web.bootstrap.grid import Container, DeviceClass


class CollapseToggle(HTMLElement):
    def __init__(self, view, target_widget, text=None, hide_for_size=None):
        text = text or '☰'

        super(CollapseToggle, self).__init__(view, 'button', children_allowed=True)
        self.set_attribute('type', 'button')

        self.target_widget = target_widget
        self.append_class('navbar-toggler')
        self.set_attribute('data-toggle', 'collapse')
        self.set_attribute('data-target', '#%s' % target_widget.css_id)
        if hide_for_size:
            self.append_class('hidden-%s-up' % hide_for_size.class_label)

        self.add_child(TextNode(view, text))


class ResponsivePull(HTMLAttributeValueOption):
    def __init__(self, side, device_class):
        is_set = True if device_class else False
        self.device_class_label = device_class if (is_set and device_class is not True) else None
        super(ResponsivePull, self).__init__(side, is_set, prefix='pull')

    @property
    def side(self):
        return self.option_string

    def as_html_snippet(self):
        if self.device_class_label:
            return '-'.join([self.prefix, self.device_class_label, self.side])
        return super(ResponsivePull, self).as_html_snippet()


class ResponsiveFloat(Layout):
    def __init__(self, left=None, right=None):
        super(ResponsiveFloat, self).__init__()
        if left and right:
            raise ProgrammerError('Both left= and right= have been given. Specify left or right, not both.')

        self.left = ResponsivePull('left', left)
        self.right = ResponsivePull('right', right)

    def customise_widget(self):
        super(ResponsiveFloat, self).customise_widget()
        for responsive_pull in [self.left, self.right]:
            if responsive_pull.is_set:
                self.widget.append_class(responsive_pull.as_html_snippet())


class NavbarFixed(HTMLAttributeValueOption):
    def __init__(self, fixed_to):
        super(NavbarFixed, self).__init__(fixed_to, fixed_to is not None, 
                                          prefix='navbar-fixed', constrain_value_to=['top', 'bottom'])

class ColourTheme(HTMLAttributeValueOption):
    def __init__(self, name):
        super(ColourTheme, self).__init__(name, name is not None, 
                                          prefix='navbar', constrain_value_to=['light', 'dark'])

class BackgroundScheme(HTMLAttributeValueOption):
    def __init__(self, name):
        super(BackgroundScheme, self).__init__(name, name is not None, 
                                               prefix='bg', constrain_value_to=['primary', 'inverse', 'faded']) 

class NavbarLayout(Layout):
    """Used to populate a Navbar.

    :keyword fixed_to: If one of 'top' or 'bottom', the Navbar will stick to the top or bottom of the viewport.
    :keyword full: If True, the Navbar fills the available width.
    :keyword center_contents: If True, all the contents of the Navbar is centered within the Navbar itself.
    :keyword colour_theme: Whether the Navbar has a 'dark' or 'light' background.
    :keyword bg_scheme: Whether the Navbar should use 'primary' colors, an 'inverse' (light on dark) scheme or a 'faded' background.
    """
    def __init__(self, fixed_to=None, full=False, center_contents=False, colour_theme=None, bg_scheme=None):
        super(NavbarLayout, self).__init__()
        if fixed_to and full:
            raise ProgrammerError('Both fixed_to and full are given. Give fixed_to or full, but not both')
        
        self.fixed = NavbarFixed(fixed_to)
        self.full = HTMLAttributeValueOption('navbar-full', full)
        self.center_contents = center_contents
        self.colour_theme = ColourTheme(colour_theme)
        self.bg_scheme = BackgroundScheme(bg_scheme)
        self.brand = None
        self.contents_container = None

    def customise_widget(self):
        super(NavbarLayout, self).customise_widget()
        nav = self.widget.html_representation
        if self.center_contents:
            centering_div = nav.add_child(Div(self.view).use_layout(Container()))
            self.contents_container = centering_div
        else:
            self.contents_container = nav

        for option in [self.fixed, self.full]:
            if option.is_set:
                self.widget.append_class(option.as_html_snippet())
            
        for option in [self.colour_theme, self.bg_scheme]:
            if option.is_set:
                nav.append_class(option.as_html_snippet())

    def set_brand_text(self, brand_text):
        """Sets the brand to be a link to the home page that contains the given text.

        :param brand_text: Text to use for branding.
        """
        brand_a = A(self.view, Url('/#'), description=brand_text)
        self.set_brand(brand_a)

    @arg_checks(brand_html_element=IsInstance(HTMLElement))
    def set_brand(self, brand_html_element):
        """Sets `brand_html_element` to be used as branding.

        :param brand_html_element: An :class:`~reahl.web.ui.HTMLElement` to be used as branding.
        """
        if self.brand:
            raise ProgrammerError('Brand has already been set to: %s' % self.brand)

        self.contents_container.insert_child(0, brand_html_element)
        brand_html_element.append_class('navbar-brand')
        self.brand = brand_html_element

    @arg_checks(widget=IsInstance((reahl.web.bootstrap.navs.Nav, Form)))
    def add(self, widget, left=None, right=None):
        """Adds the given Form or Nav `widget` to the Navbar.

        :param widget: A :class:`~reahl.web.bootstrap.navs.Nav` or :class:`~reahl.web.bootstrap.ui.Form` to add.
        :keyword left: If True, `widget` is aligned to the left of the Navbar.
        :keyword right: If True, `widget` is aligned to the right of the Navbar.
        """
        if isinstance(widget, reahl.web.bootstrap.navs.Nav):
            widget.append_class('navbar-nav')
        if left or right:
            child = Div(self.view).use_layout(ResponsiveFloat(left=left, right=right))
            child.add_child(widget)
        else:
            child = widget
        self.contents_container.add_child(child)
        return widget

    def add_toggle(self, target_html_element, text=None):
        """Adds a link that toggles the display of the given `target_html_element`.

        :param target_html_element: A :class:`~reahl.web.ui.HTMLElement`
        :keyword text: Text to be used on the toggle link.
        """
        if not target_html_element.css_id_is_set:
            raise ProgrammerError('%s has no css_id set. A toggle is required to have a css_id' % target_html_element)
        target_html_element.append_class('collapse')
        toggle = CollapseToggle(self.view, target_html_element, text=text)
        self.contents_container.add_child(toggle)
        return toggle


class ResponsiveLayout(NavbarLayout):
    def __init__(self, collapse_below_device_class, fixed_to=None, full=False, center_contents=False, colour_theme=None, bg_scheme=None, text=None):
        super(ResponsiveLayout, self).__init__(fixed_to=fixed_to, full=full, center_contents=center_contents, colour_theme=colour_theme, bg_scheme=bg_scheme)
        self.collapse_below_device_class = DeviceClass(collapse_below_device_class)
        if not self.collapse_below_device_class.one_smaller:
            raise ProgrammerError(('There is no device class smaller than %s' % self.collapse_below_device_class)+\
                                  ' It does not make sense to collapse only if the viewport is smaller than the smallest device')
        self.text = text

    def customise_widget(self):
        super(ResponsiveLayout, self).customise_widget()
        if not self.widget.css_id_is_set:
            raise ProgrammerError('%s has no css_id set. A %s can only be used with a Widget that has a css_id' % \
                                (self.widget, self.__class__))

        collapsable = Div(self.view, css_id='%s_collapsable' % self.widget.css_id)
        collapsable.append_class('collapse')
        toggle_widget = CollapseToggle(self.view, collapsable, text=self.text, hide_for_size=self.collapse_below_device_class)
        toggle_size = self.collapse_below_device_class.one_smaller
        collapsable.append_class('navbar-toggleable-%s' % toggle_size.class_label)
        
        self.contents_container.add_child(toggle_widget)
        self.contents_container.add_child(collapsable)
        self.contents_container = collapsable


class Navbar(HTMLWidget):
    """A bootstrap Navbar can be used as page header in a web
    application. A Navbar can contain some branding text, a
    :class:`~reahl.web.bootstrap.navs.Nav` and a small
    :class:`~reahl.web.bootstrap.forms.Form`.

    You populate a Navbar using a NavbarLayout for complete control
    over how the Navbar is laid out, which of its possible elements 
    are present and how they are composed themselves.

    :param view: (See :class:`reahl.web.fw.Widget`)
    """
    def __init__(self, view):
        super(Navbar, self).__init__(view)

        self.navbar = self.add_child(Nav(view))
        self.navbar.append_class('navbar')
        self.set_html_representation(self.navbar)








from setuptools import setup, Command
from pkg_resources import require
class InstallTestDependencies(Command):
    user_options = []
    def run(self):
        from setuptools.command import easy_install
        if self.distribution.tests_require: easy_install.main(self.distribution.tests_require)

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

setup(
    name='reahl-web',
    version='3.2.0',
    description='The core Reahl web framework',
    long_description='Reahl is a web application framework that allows a Python programmer to work in terms of useful abstractions - using a single programming language.\n\nThis package contains the core of the Reahl framework.\n\nSee http://www.reahl.org/docs/3.2/tutorial/gettingstarted.d.html for installation instructions. ',
    url='http://www.reahl.org',
    maintainer='Iwan Vosloo',
    maintainer_email='iwan@reahl.org',
    packages=['reahl', 'reahl.web', 'reahl.messages', 'reahl.web_dev', 'reahl.web.static', 'reahl.web.holder', 'reahl.web.bootstrap', 'reahl.web.static.jquery', 'reahl.web_dev.appstructure', 'reahl.web_dev.advanced', 'reahl.web_dev.widgets', 'reahl.web_dev.inputandvalidation', 'reahl.web_dev.bootstrap', 'reahl.web_dev.advanced.subresources'],
    py_modules=['setup'],
    include_package_data=True,
    package_data={'': ['*/LC_MESSAGES/*.mo']},
    namespace_packages=['reahl', 'reahl.messages'],
    install_requires=['reahl-component>=3.2,<3.3', 'ply>=3.8,<3.8.999', 'slimit>=0.8,<0.8.999', 'cssmin>=0.2,<0.2.999', 'BeautifulSoup4>=4.3,<4.3.999', 'webob>=1.4,<1.4.999', 'Babel>=2.1,<2.1.999'],
    setup_requires=['reahl-bzrsupport>=3.2,<3.3'],
    tests_require=['nose', 'reahl-tofu>=3.2,<3.3', 'reahl-stubble>=3.2,<3.3', 'reahl-sqlalchemysupport>=3.2,<3.3', 'reahl-postgresqlsupport>=3.2,<3.3', 'reahl-web-declarative>=3.2,<3.3', 'reahl-domain>=3.2,<3.3', 'reahl-webdev>=3.2,<3.3', 'reahl-dev>=3.2,<3.3'],
    test_suite='reahl.web_dev',
    entry_points={
        'reahl.attachments.any': [
            '0-reahl+web+static+runningon.png = reahl'    ],
        'reahl.translations': [
            'reahl-web = reahl.messages'    ],
        'reahl.attachments.css': [
            '0-reahl+web+reahl.labelledinput.css = reahl',
            '1-reahl+web+reahl.labeloverinput.css = reahl',
            '2-reahl+web+reahl.menu.css = reahl',
            '3-reahl+web+reahl.hmenu.css = reahl',
            '4-reahl+web+reahl.slidingpanel.css = reahl'    ],
        'reahl.attachments.js': [
            '0-reahl+web+reahl.form.js = reahl',
            '1-reahl+web+reahl.cueinput.js = reahl',
            '2-reahl+web+reahl.labeloverinput.js = reahl',
            '3-reahl+web+reahl.fileuploadli.js = reahl',
            '4-reahl+web+reahl.fileuploadpanel.js = reahl',
            '5-reahl+web+reahl.popupa.js = reahl',
            '6-reahl+web+reahl.slidingpanel.js = reahl'    ],
        'reahl.eggs': [
            'Egg = reahl.component.eggs:ReahlEgg'    ],
        'reahl.configspec': [
            'config = reahl.web.egg:WebConfig'    ],
                 },
    extras_require={},
    cmdclass={'install_test_dependencies': InstallTestDependencies}
)



# Copyright 2014, 2015, 2016 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import print_function, unicode_literals, absolute_import, division

from reahl.stubble import EmptyStub
from reahl.tofu import Fixture, test, scenario
from reahl.tofu import vassert

from reahl.webdev.tools import XPath, WidgetTester
from reahl.web.ui import StaticColumn, DynamicColumn, Table, Thead, Span, Div, P
from reahl.web_dev.fixtures import WebBasicsMixin, WebFixture

from reahl.component.modelinterface import Field, BooleanField


class DataItem(object):
    def __init__(self, row, alpha):
        self.row = row
        self.alpha = alpha


class TableFixture(Fixture, WebBasicsMixin):
    def new_data(self):
        return [DataItem(1, 'T'), DataItem(2, 'H'), DataItem(3, 'E')]

    def table_caption_is(self, expected):
        return  self.driver_browser.find_element(XPath.caption_with_text(expected))

    def table_summary_is(self, summary):
        return  self.driver_browser.find_element(XPath.table_with_summary(summary))

    def table_column_name_is(self, column_number, expected):
        return self.driver_browser.web_driver.find_element_by_xpath('((//table/thead/tr/th)[%s]/span)[1]' % column_number).text == expected

    def get_table_row(self, row_number):
        row_data = []
        for column_number in list(range(1,self.table_number_columns()+1)):
            row_data.append(self.driver_browser.web_driver.find_element_by_xpath('((//table/tbody/tr)[%s]/td)[%s]' % (row_number, column_number)).text)
        return row_data

    def table_number_columns(self):
        return len(self.driver_browser.web_driver.find_elements_by_xpath('//table/thead/tr/th'))

    def table_number_rows(self):
        return len(self.driver_browser.web_driver.find_elements_by_xpath('//table/tbody/tr'))


@test(TableFixture)
def table_basics(fixture):
    """A Table created .from_columns() displays a list of items as defined by a list of Columns"""

    class MainWidget(Div):
        def __init__(self, view):
            super(MainWidget, self).__init__(view)
            table = Table.from_columns(view, 
                            [StaticColumn(Field(label='Row Number'), 'row'),
                             StaticColumn(Field(label='Alpha'), 'alpha')],
                            fixture.data,
                            caption_text='All my friends',
                            summary='Summary for screen reader',
                            css_id='my_table_data'
                            )
            self.add_child(table)

    wsgi_app = fixture.new_wsgi_app(enable_js=True, child_factory=MainWidget.factory())
    fixture.reahl_server.set_app(wsgi_app)
    fixture.driver_browser.open('/')
        
    # The table has a caption and summary
    vassert( fixture.table_caption_is('All my friends') )
    vassert( fixture.table_summary_is('Summary for screen reader') )

    # Column headings are derived from given Column Fields
    vassert( fixture.table_column_name_is(1, 'Row Number') )
    vassert( fixture.table_column_name_is(2, 'Alpha') )

    # A string representation of the value of each Field of a given data item is shown in the appropriate cell
    vassert( fixture.table_number_rows() == 3 )

    vassert( fixture.get_table_row(1) == ['1' ,'T'] )
    vassert( fixture.get_table_row(2) == ['2' ,'H'] )
    vassert( fixture.get_table_row(3) == ['3' ,'E'] )


class ColumnFixture(WebFixture):
    sort_key = EmptyStub()
    heading = 'A heading'
    row_item = EmptyStub(some_attribute=True)
    
    @scenario
    def static_column(self):
        """StaticColumn represents an attribute of the item in a row, using a Field to translate the value of that attribute into a string and to specify a column header."""
        self.column = StaticColumn(BooleanField(label=self.heading), 'some_attribute', sort_key=self.sort_key)
        self.expected_cell_html = 'on' # as translated by BooleanField
        self.expected_heading_html = '<span>A heading</span>'

    @scenario
    def dynamic_column(self):
        """DynamicColumn uses a function to create a Widget on the fly for each cell and uses the specified heading."""
        def make_span(view, data_item):
            return Span(view, text='Answer: %s' % (data_item.some_attribute))

        self.column = DynamicColumn(self.heading, make_span, sort_key=self.sort_key)
        self.expected_cell_html = '<span>Answer: True</span>' # raw attribute used
        self.expected_heading_html = '<span>A heading</span>'

    @scenario
    def dynamic_column_static_heading(self):
        """A DynamicColumn can also use a function to create its heading on the fly."""
        def make_span(view, data_item):
            return Span(view, text='Answer: %s' % (data_item.some_attribute))

        def make_heading(view):
            return P(view, text=self.heading)

        self.column = DynamicColumn(make_heading, make_span, sort_key=self.sort_key)
        self.expected_cell_html = '<span>Answer: True</span>' # raw attribute used
        self.expected_heading_html = '<p>A heading</p>'


@test(ColumnFixture)
def different_kinds_of_columns(fixture):
    """There are different kinds of Columns, allowing different levels of flexibility for defining a Table"""

    vassert( fixture.column.sort_key is fixture.sort_key )

    # The heading
    widget_for_heading = fixture.column.heading_as_widget(fixture.view)
    actual = WidgetTester(widget_for_heading).render_html()

    vassert( actual == fixture.expected_heading_html )

    # A cell
    widget_for_cell = fixture.column.as_widget(fixture.view, fixture.row_item)
    actual = WidgetTester(widget_for_cell).render_html()

    vassert( actual == fixture.expected_cell_html )


@test(TableFixture)
def table_thead(fixture):
    """Table can find its Thead element"""

    table = Table(fixture.view)
    thead = table.add_child(Thead(fixture.view))

    vassert( table.thead is thead )




# Copyright 2016 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
.. versionadded:: 3.2


Styled Inputs that allow a user to choose or upload files.

"""
from __future__ import print_function, unicode_literals, absolute_import, division

import six

from reahl.component.i18n import Translator
from reahl.component.modelinterface import exposed, Action, Event, Field, UploadedFile
from reahl.web.fw import WebExecutionContext
import reahl.web.ui
from reahl.web.ui import PrimitiveInput, UniqueFilesConstraint
from reahl.web.bootstrap.ui import Div, Span, Li, Ul
from reahl.web.bootstrap.forms import Button, NestedForm, ButtonLayout, FormLayout, Label


_ = Translator('reahl-web')



class _UnstyledHTMLFileInput(reahl.web.ui.SimpleFileInput):
    append_error = False
    add_default_attribute_source = False

    def __init__(self, form, bound_field):
        super(_UnstyledHTMLFileInput, self).__init__(form, bound_field)
        self.append_class('form-control-file')


class FileInputButton(reahl.web.ui.WrappedInput):
    """A single button which activated the browser's file choice dialog
       when clicked. The chosen file will only be uploaded once the
       user clicks on any :class:`Button` associated with the same
       :class:`Form` as this Input.

       :param form: (See :class:`~reahl.web.ui.Input`)
       :param bound_field: (See :class:`~reahl.web.ui.Input`, must be of 
              type :class:`reahl.component.modelinterface.FileField`)

    """
    def __init__(self, form, bound_field):
        label = Label(form.view)
        self.simple_input = label.add_child(_UnstyledHTMLFileInput(form, bound_field))
        self.simple_input.html_representation.append_class('btn-secondary')
        label.add_child(Span(form.view, text=_('Choose file(s)')))
        super(FileInputButton, self).__init__(self.simple_input)
        self.add_child(label)

        label.append_class('reahl-bootstrapfileinputbutton')
        label.append_class('btn')
        label.append_class('btn-primary')
        self.set_html_representation(label)

    @property
    def html_control(self):
        return self.simple_input.html_control

    def get_js(self, context=None):
        js = ['$(".reahl-bootstrapfileinputbutton").bootstrapfileinputbutton({});']
        return super(FileInputButton, self).get_js(context=context) + js


class FileInput(reahl.web.ui.WrappedInput):
    """A visual combination of a two buttons and a status area. When the
    user clicks on the 'Choose file' button, the browser's file choice
    dialog is activated. Once chosen the file name that was chosen is
    shown in the status area. The last button will upload this file
    when clicked (it is automatically clicked is the user's JavaScript
    is enabled).

    :param form: (See :class:`~reahl.web.ui.Input`)
    :param bound_field: (See :class:`~reahl.web.ui.Input`, must be of 
           type :class:`reahl.component.modelinterface.FileField`)

    """
    def __init__(self, form, bound_field):
        file_input = FileInputButton(form, bound_field)
        super(FileInput, self).__init__(file_input)

        self.input_group = self.add_child(Div(self.view))
        self.input_group.append_class('input-group')
        self.input_group.append_class('reahl-bootstrapfileinput')
        self.set_html_representation(self.input_group)

        span = self.input_group.add_child(Span(form.view))
        span.append_class('input-group-btn')
        span.add_child(file_input)

        filename_input = self.input_group.add_child(Span(self.view, text=_('No files chosen')))
        filename_input.append_class('form-control')


    def get_js(self, context=None):
        js = ['$(".reahl-bootstrapfileinput").bootstrapfileinput({nfilesMessage: "%s", nofilesMessage: "%s"});' % \
              (_('files chosen'), _('No files chosen'))]
        return super(FileInput, self).get_js(context=context) + js



# Uses: reahl/web/reahl.files.js
class FileUploadLi(Li):
    def __init__(self, form, remove_event, persisted_file, css_id=None):
        super(FileUploadLi, self).__init__(form.view, css_id=css_id)
        self.set_attribute('class', 'reahl-bootstrap-file-upload-li')
        self.add_child(Button(form, remove_event.with_arguments(filename=persisted_file.filename)))
        self.add_child(Span(self.view, persisted_file.filename))

    def get_js(self, context=None):
        return ['$(".reahl-bootstrap-file-upload-li").bootstrapfileuploadli();']



# Uses: reahl/web/reahl.files.js
class FileUploadPanel(Div):
    def __init__(self, file_upload_input, css_id=None):
        super(FileUploadPanel, self).__init__(file_upload_input.view, css_id=css_id)
        self.set_attribute('class', 'reahl-bootstrap-file-upload-panel')
        self.file_upload_input = file_upload_input

        self.add_nested_form()
        self.add_uploaded_list()
        self.add_upload_controls()

    def add_nested_form(self):
        self.upload_form = self.add_child(NestedForm(self.view, '%s-%s-upload' % (self.input_form.css_id, self.bound_field.name)))
        self.upload_form.define_event_handler(self.events.upload_file)
        self.upload_form.define_event_handler(self.events.remove_file)
        self.upload_form.form.set_attribute('novalidate','novalidate')

    @property
    def name(self):
        return self.bound_field.variable_name

    def add_uploaded_list(self):
        ul = self.upload_form.add_child(Ul(self.view))
        for persisted_file in self.persisted_file_class.get_persisted_for_form(self.input_form, self.name):
            ul.add_child(FileUploadLi(self.upload_form.form, self.events.remove_file, persisted_file))

    def add_upload_controls(self):
        controls_panel = self.upload_form.add_child(Div(self.view)).use_layout(FormLayout())
        file_input = controls_panel.layout.add_input(FileInput(self.upload_form.form, self.fields.uploaded_file), hide_label=True)

        button_addon = file_input.html_representation.add_child(Span(self.view))
        button_addon.append_class('input-group-btn')
        button_addon.add_child(Button(self.upload_form.form, self.events.upload_file))


    @property
    def persisted_file_class(self):
        return self.file_upload_input.persisted_file_class

    @property
    def input_form(self):
        return self.file_upload_input.form

    @property
    def bound_field(self):
        return self.file_upload_input.bound_field

    @exposed
    def events(self, events):
        events.upload_file = Event(label=_('Upload'), action=Action(self.upload_file))
        events.remove_file = Event(label=_('Remove'),
                                   action=Action(self.remove_file, ['filename']),
                                   filename=Field(required=True))

    @exposed
    def fields(self, fields):
        fields.uploaded_file = self.bound_field.unbound_copy()
        fields.uploaded_file.disallow_multiple()
        fields.uploaded_file.label = _('Add file')
        fields.uploaded_file.make_optional()
        fields.uploaded_file.add_validation_constraint(UniqueFilesConstraint(self.input_form, self.bound_field.name))

    def attach_jq_widget(self, selector, widget_name, **options):
        def js_repr(value):
            if isinstance(value, six.string_types):
                return '"%s"' % value
            return value
        option_args = ','.join(['%s: %s' % (name, js_repr(value)) for name, value in options.items()])
        return '$(%s).%s({%s});' % (selector, widget_name, option_args)

    def get_js(self, context=None):
        selector = self.contextualise_selector('"#%s .reahl-bootstrap-file-upload-panel"' % self.input_form.css_id, context)
        unique_names_constraint = self.fields.uploaded_file.get_validation_constraint_of_class(UniqueFilesConstraint)
        js = self.attach_jq_widget(selector, 'bootstrapfileuploadpanel',
                    form_id=self.upload_form.form.css_id,
                    nested_form_id=self.upload_form.css_id,
                    input_form_id=self.input_form.css_id,
                    errorMessage=_('an error occurred, please try again later.'),
                    removeLabel=self.events.remove_file.label,
                    cancelLabel=_('Cancel'),
                    duplicateValidationErrorMessage=unique_names_constraint.message,
                    waitForUploadsMessage=_('Please try again when all files have finished uploading.'))
        return super(FileUploadPanel, self).get_js(context=context) + [js]

    def remove_file(self, filename):
        self.persisted_file_class.remove_persisted_for_form(self.input_form, self.name, filename)

    def upload_file(self):
        if self.uploaded_file is not None:
            self.persisted_file_class.add_persisted_for_form(self.input_form, self.name, self.uploaded_file)




class FileUploadInput(PrimitiveInput):
    """A Widget that allows the user to upload several files.  FileUploadInput
    makes use of JavaScript to save a user some time: once you choose a file, 
    it is immediately uploaded to the server in the background so that you can
    continue choosing more files.

    Controls are provided so you can cancel uploads that are in
    progress or remove ones that have finished. While a file is
    uploading a progress bar is also shown.

    :param form: (See :class:`~reahl.web.ui.Input`)
    :param bound_field: (See :class:`~reahl.web.ui.Input`, must be of 
              type :class:`reahl.component.modelinterface.FileField`)
    """
    append_error = False
    add_default_attribute_source = False

    @property
    def persisted_file_class(self):
        config = WebExecutionContext.get_context().config
        return config.web.persisted_file_class

    def create_html_widget(self):
        return FileUploadPanel(self)

    @property
    def html_control(self):
        return None

    def get_value_from_input(self, input_values):
        return [UploadedFile(f.filename, f.file_obj.read(), f.mime_type)
                 for f in self.persisted_file_class.get_persisted_for_form(self.form, self.name)]

    def enter_value(self, input_value):
        pass

